﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Reflection;
using System.Threading;

namespace jello
{
    class Program
    {
        private static string help = "Useage: jello [flags] [options] file_name\nFlags:\n  -v: Displays version\n  -h: Displays this help section\nOptions:\n  -dbl: Sets the debug level, can be 0 to 2\n  -t: Sets internal clock rate in hertz";
        private static string versionNumber = Assembly.GetExecutingAssembly().GetName().Version.Major.ToString() + "." + Assembly.GetExecutingAssembly().GetName().Version.Minor.ToString() + "." + Assembly.GetExecutingAssembly().GetName().Version.Build.ToString();
        private static string versionName = "Jellal"; //Because everything is Jellal
        private static string version = "JELLO v" + versionNumber + " (" + versionName + ") by Jacob Allen";
        private static string[] validFlags = { "-v", "-h", "-jellal" };
        private static Commands commandSet = new Commands();

        static void Main(string[] args)
        {
            int printLevel = 0;
            float tickRate = 0;

            if (args.Length < 1)
            {
                Console.WriteLine("Command requires arguments.");
                Environment.Exit(1);
            }

            Dictionary<string, string> options = ProcessArgs(args);

            if (options.Keys.Contains("-jellal"))
            {
                Console.WriteLine(version);
                Console.WriteLine("Becuase everything is Jellal.");
                Environment.Exit(0);
            }

            if (options.Keys.Contains("-h"))
            {
                Console.WriteLine(version);
                Console.WriteLine(help);
                Environment.Exit(0);
            }

            if (options.Keys.Contains("-v"))
            {
                Console.WriteLine(version);
                Environment.Exit(0);
            }

            if (options.Keys.Contains("-dbl"))
            {
                int dbl = 0;
                if (int.TryParse(options["-dbl"], out dbl))
                {
                    printLevel = dbl;
                }
                else
                {
                    Console.WriteLine("Debug level must be an integer.");
                    Environment.Exit(1);
                }
            }

            if (options.Keys.Contains("-t"))
            {
                float tick = 0;
                if (float.TryParse(options["-t"], out tick))
                {
                    tickRate = tick;
                }
                else
                {
                    Console.WriteLine("Clock speed must be in hertz.");
                    Environment.Exit(1);
                }
            }

            if (!options.Keys.Contains("*filename*"))
            {
                Console.WriteLine("A file name is required.");
                Environment.Exit(1);
            }

            Console.WriteLine(version);

            Console.Write("Reading File... ");

            List<string> codelines = LoadFile(options["*filename*"]);

            Console.WriteLine("Done.");

            Console.WriteLine("Processing Commands...");

            RunCode(codelines, tickRate, printLevel);

            Console.WriteLine("Done.");
        }

        private static Dictionary<string, string> ProcessArgs(string[] args)
        {
            List<string> toProcess = args.ToList();
            Dictionary<string, string> options = new Dictionary<string, string> { };

            while (toProcess.Count > 0)
            {
                string item = toProcess[0];
                toProcess.RemoveAt(0);
                if (item[0] == '-')
                {
                    if (validFlags.Contains(item))
                    {
                        options[item] = "true";
                    }
                    else
                    {
                        if (toProcess.Count > 0)
                        {
                            options[item] = toProcess[0];
                            toProcess.RemoveAt(0);
                        }
                        else
                        {
                            Console.WriteLine("Option requires value.");
                            Environment.Exit(1);
                        }
                    }

                }
                else
                {
                    if (options.Keys.Contains("*filename*"))
                    {
                        Console.WriteLine("Only one file name is required.");
                        Environment.Exit(1);
                    }
                    else
                    {
                        options["*filename*"] = item;
                    }
                }
            }

            return options;
        }

        private static List<string> LoadFile(string filename)
        {
            List<string> codelines = new List<string> { };
            string line = "";
            try
            {
                StreamReader file = new StreamReader(filename);
                while ((line = file.ReadLine()) != null)
                {
                    codelines.Add(line);
                }
                file.Close();
            }
            catch (IOException)
            {
                Console.WriteLine("Error while reading file '" + filename + "'.");
                Environment.Exit(1);
            }
            return codelines;
        }

        private static bool RunCode(List<string> codelines, float tickRate, int debugLevel)
        {
            InternalState State = new InternalState();
            State.Pointer = 0;
            State.CodeLines = codelines;
            int previousPointer = 0;

            while (State.Pointer < State.CodeLines.Count)
            {
                string line = State.CodeLines[State.Pointer].TrimStart(new char[] { ' ', '\n', '\t', '\r' });

                if (line == "" || line[0] == '#' || line[0] == ':')
                {
                    State.Pointer += 1;
                    continue;
                }

                List<string> parts = line.Split(' ').ToList();
                string command = parts[0];
                parts.RemoveAt(0);
                string[] values = parts.ToArray();

                if (debugLevel > 0)
                    Console.WriteLine("  -> " + line + " AT " + (State.Pointer + 1).ToString());

                previousPointer = State.Pointer;

                int pointerMove = commandSet.RunCommand(command, values, ref State);

                State.Pointer += pointerMove;

                if (State.CleanUp())
                {
                    Console.WriteLine("Error on line " + (previousPointer+1).ToString() + " at " + codelines[previousPointer]);
                    Console.WriteLine(State.Dump());
                    Console.WriteLine("Internal state error, program terminating.");
                    Environment.Exit(1);
                }

                if (debugLevel > 1)
                    Console.WriteLine(State.Dump());

                if (tickRate > 0)
                    Thread.Sleep((int)((1.0f / tickRate) * 1000));
            }
            return true;
        }
    }
}
